/*
	hufdec.c - Huffman decoder
	(c) 2006  Avelino Herrera Morales

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

	info:
		http://articulos.conclase.net/compresion/huffman.html
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>



#define  CODE_SIZE_BYTES  80

typedef struct _t_node {
	unsigned char terminal;
	int freq;
	int data;
	struct _t_node *left_node;
	struct _t_node *right_node;
	struct _t_node *parent_node;
	unsigned char code[CODE_SIZE_BYTES];
	int code_size_bits;
} t_node;

typedef struct _t_file_node {
	unsigned char data;
	unsigned short left_node_offset;
	unsigned short right_node_offset;
} __attribute__ ((__packed__)) t_file_node;


t_node **list;
int list_size = 0;


t_node *read_node(int fd, unsigned short offset) {
	t_file_node file_node;
	t_node *node;

	lseek(fd, offset, SEEK_SET);
	read(fd, &file_node, sizeof(t_file_node));
	node = (t_node *) malloc(sizeof(t_node));
	if (file_node.left_node_offset > 0)
		node->left_node = read_node(fd, file_node.left_node_offset);
	else
		node->left_node = NULL;
	if (file_node.right_node_offset > 0)
		node->right_node = read_node(fd, file_node.right_node_offset);
	else
		node->right_node = NULL;
	if ((node->left_node != NULL) || (node->right_node != NULL)) {
		node->terminal = 0;
		node->data = 0;
	}
	else {
		node->terminal = 1;
		node->data = file_node.data;
	}
	return node;
}

t_node *read_tree(int fd) {
	unsigned short root_node_offset;

	// read root node offset
	lseek(fd, 3, SEEK_SET);
	read(fd, &root_node_offset, sizeof(unsigned short));
	// goto root node offset
	return read_node(fd, root_node_offset);
}

void uncompress(int input_fd, t_node *root, int output_fd) {
	unsigned short byte_offset;
	unsigned short bit_offset;
	unsigned char current_byte;
	unsigned char current_bit;
	t_node *current_node;
	int readed, compressed_size_bits, offset_bits;

	// read data offset
	lseek(input_fd, 5, SEEK_SET);
	read(input_fd, &byte_offset, 2);
	// read compressed size in bits
	lseek(input_fd, byte_offset, SEEK_SET);
	read(input_fd, &compressed_size_bits, sizeof(int));
	printf("compressed_size_bits = %d (at %04x)\n", compressed_size_bits, byte_offset);
	// read first byte
	printf("data starts at %04x\n", lseek(input_fd, 0, SEEK_CUR));
	readed = read(input_fd, &current_byte, 1);
	// init
	bit_offset = 0;
	offset_bits = 0;
	current_node = root;
	while (readed == 1) {
		if (bit_offset > 7) {
			bit_offset = 0;
			byte_offset++;
			//lseek(input_fd, byte_offset, SEEK_SET);
			readed = read(input_fd, &current_byte, 1);
		}
		current_bit = (current_byte >> bit_offset) & 1;
		if (current_bit) {
			current_node = current_node->right_node;
			//printf("1");
		}
		else {
			current_node = current_node->left_node;
			//printf("0");
		}
		if (current_node->terminal) {
			// emit current_node->data
			unsigned char byte_data = (unsigned char) current_node->data;
			write(output_fd, &byte_data, 1);
			current_node = root;
		}
		bit_offset++;
		offset_bits++;
		if (offset_bits >= compressed_size_bits)
			break;
	}
}

void show_node(int level, t_node *node) {
	int i;
	char buffer[80] = "";

	if (node == NULL)
		return;
	for (i = 0; i < level; i++)
		printf("   ");
	/* get_code_string(node, buffer); */
	if (node->data < 0)
		printf(" *  (%d) [%s]\n", node->freq, buffer);
	else {
		if (isgraph(node->data))
			printf("'%c' (%d) [%s]\n", node->data, node->freq, buffer);
		else
			printf("%02x  (%d) [%s]\n", node->data, node->freq, buffer);
	}
	show_node(level + 1, node->left_node);
	show_node(level + 1, node->right_node);
}

int main(int argc, char *argv[]) {
	int input_fd;
	int output_fd;
	t_node *root;

	if (argc < 3) {
		printf("syntax: %s source_file destination_file\n", argv[0]);
		return -1;
	}
	input_fd = open(argv[1], O_RDONLY);
	if (input_fd < 0) {
		perror("open compressed input file");
		return -1;
	}
	output_fd = open(argv[2], O_WRONLY | O_CREAT | O_TRUNC, 0644);
	if (output_fd < 0) {
		perror("open output file");
		return -1;
	}
	root = read_tree(input_fd);
	show_node(0, root);
	uncompress(input_fd, root, output_fd);
	close(input_fd);
	close(output_fd);
	return 0;
}

