/*
	hufenc.c - Huffman encoder
	(c) 2006  Avelino Herrera Morales

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

	info:
		http://articulos.conclase.net/compresion/huffman.html
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


/*


*/

#define  CODE_SIZE_BYTES  80

typedef struct _t_node {
	int freq;
	int data;
	struct _t_node *left_node;
	struct _t_node *right_node;
	struct _t_node *parent_node;
	unsigned char code[CODE_SIZE_BYTES];
	int code_size_bits;
} t_node;

typedef struct _t_file_node_header {
	unsigned char data;
	unsigned short left_node_offset;
	unsigned short right_node_offset;
} __attribute__ ((__packed__)) t_file_node_header;

typedef struct _t_output_bit_stream {
	int bit_position;
	unsigned char byte;
	unsigned char *buffer;
	int buffer_size;
} t_output_bit_stream;


t_node **list;
int list_size = 0;


t_output_bit_stream *output_bit_stream_init(int size) {
	t_output_bit_stream *ret = (t_output_bit_stream *) malloc(sizeof(t_output_bit_stream));
	ret->bit_position = 0;
	ret->byte = 0;
	ret->buffer = (unsigned char *) malloc(size);
	ret->buffer_size = 0;
	return ret;
}

void output_bit_stream_send(t_output_bit_stream *stream, unsigned char bit) {
	bit <<= stream->bit_position;
	stream->byte |= bit;
	if (stream->bit_position == 7) {
		stream->buffer[stream->buffer_size] = stream->byte;
		stream->buffer_size++;
		stream->byte = 0;
		stream->bit_position = 0;
	}
	else
		stream->bit_position++;
}

void output_bit_stream_end(t_output_bit_stream *stream) {
	if (stream->bit_position > 0) {
		stream->buffer[stream->buffer_size] = stream->byte;
		stream->buffer_size++;
	}
	/* printf("bit_stream: %d byte(s)\n", stream->buffer_size); */
}

int get_code(t_node *root, unsigned char data, unsigned char *dest, int *dest_size_bits) {
	unsigned char left_code[CODE_SIZE_BYTES];
	unsigned char right_code[CODE_SIZE_BYTES];
	int left_code_size_bits;
	int right_code_size_bits;
	int ret;

	if (root == NULL)
		return 0;
	if (root->data == data) {
		memcpy(dest, root->code, CODE_SIZE_BYTES);
		*dest_size_bits = root->code_size_bits;
		return 1;
	}
	ret = get_code(root->left_node, data, left_code, &left_code_size_bits);
	if (ret == 0) {
		ret = get_code(root->right_node, data, right_code, &right_code_size_bits);
		memcpy(dest, right_code, CODE_SIZE_BYTES);
		*dest_size_bits = right_code_size_bits;
		return ret;
	}
	else {
		memcpy(dest, left_code, CODE_SIZE_BYTES);
		*dest_size_bits = left_code_size_bits;
		return ret;
	}
}

void compress(t_node *root, unsigned char *uncompressed_data, int uncompressed_size, unsigned char **compressed_data, int *compressed_size_bits) {
	int i, j, n;
	t_output_bit_stream *stream;
	unsigned char code[CODE_SIZE_BYTES];;
	int code_size_bits;
	int byte_pos, bit_pos;

	stream = output_bit_stream_init(uncompressed_size);
	for (i = 0; i < uncompressed_size; i++) {
		get_code(root, uncompressed_data[i], code, &code_size_bits);
		for (j = 0; j < code_size_bits; j++) {
			byte_pos = (j >> 3);
			bit_pos = (j & 7);
			if (code[byte_pos] & (1 << bit_pos))
				output_bit_stream_send(stream, 1);
			else
				output_bit_stream_send(stream, 0);
		}
	}
	output_bit_stream_end(stream);
	*compressed_data = (unsigned char *) malloc(stream->buffer_size);
	memcpy(*compressed_data, stream->buffer, stream->buffer_size);
	//*compressed_size = stream->buffer_size;
	*compressed_size_bits = ((stream->buffer_size - 1) * 8) + stream->bit_position;
}

t_node *lookup_node(unsigned char data) {
	int i;
	t_node *aux;

	for (i = 0; i < list_size; i++) {
		aux = list[i];
		if (aux->data == data)
			return aux;
	}
	list_size++;
	list = (t_node **) realloc(list, list_size * sizeof(t_node *));
	aux = (t_node *) malloc(sizeof(t_node));
	aux->data = data;
	aux->freq = 0;
	aux->left_node = NULL;
	aux->right_node = NULL;
	aux->parent_node = NULL;
	aux->code_size_bits = 0;
	memset(aux->code, 0, CODE_SIZE_BYTES);
	list[list_size - 1] = aux;
	return aux;
}

void add_node(t_node *node) {
	list_size++;
	list = (t_node **) realloc(list, list_size * sizeof(t_node *));
	list[list_size - 1] = node;
}

void delete_first_two_nodes() {
	memmove(list, &(list[2]), (list_size - 2) * sizeof(t_node *));
	list_size -= 2;
	list = (t_node **) realloc(list, list_size * sizeof(t_node *));
}

void read_list(unsigned char *source_data, int size) {
	int i;
	unsigned char buffer;
	t_node *aux;

	for (i = 0; i < size; i++) {
		aux = lookup_node(source_data[i]);
		aux->freq++;
	}
}

void order_list(int left, int right) {
	t_node *temp_node;
	int i = left;
	int j = right;
	int pivote = list[(left + right) / 2]->freq;

	while (i <= j) {
		while (list[i]->freq < pivote)
			i++;
		while (list[j]->freq > pivote)
			j--;
		if (i <= j) {
			temp_node = list[i];
			list[i] = list[j];
			list[j] = temp_node;
			i++;
			j--;
		}
	}
	if (left < j)
		order_list(left, j);
	if (i < right)
		order_list(i, right);
}

t_node *join_first_two_nodes() {
	t_node *new_node;
	t_node *left;
	t_node *right;
	left = list[0];
	right = list[1];
	delete_first_two_nodes();
	new_node = (t_node *) malloc(sizeof(t_node));
	new_node->data = -1;
	new_node->freq = left->freq + right->freq;
	new_node->left_node = left;
	new_node->right_node = right;
	left->parent_node = new_node;
	right->parent_node = new_node;
	return new_node;
}

void get_code_string(t_node *node, char *dest) {
	int i;
	int byte_position;
	int bit_position;

	dest[0] = 0;
	for (i = 0; i < node->code_size_bits; i++) {
		byte_position = i >> 3;
		bit_position = i & 7;
		if ((node->code[byte_position] >> bit_position) & 1)
			dest[i] = '1';
		else
			dest[i] = '0';
	}
	dest[i] = 0;
}

void tabulate(int tab) {
	int i;

	for (i = 0; i < tab; i++)
		printf("\t");
}

void code_node(int right, unsigned char *parent_code, int parent_code_size_bits, t_node *node) {
	unsigned int byte_position;
	unsigned int bit_position;
	/* static char aux_str[80]; */

	/* tabulate(tab); */
	if (node == NULL) {
		/* printf("null node\n"); */
		return;
	}
	memcpy(node->code, parent_code, CODE_SIZE_BYTES);
	node->code_size_bits = parent_code_size_bits;
	/*
	get_code_string(node, aux_str);
	printf("node code before concatenate bit: '%s' (size_bits=%d)\n", aux_str, node->code_size_bits);
	tabulate(tab);
	*/
	byte_position = node->code_size_bits >> 3;
	bit_position = node->code_size_bits & 7;
	if (right)
		node->code[byte_position] |= ((unsigned char)1 << bit_position);
	else
		node->code[byte_position] &= ~((unsigned char)1 << bit_position);
	node->code_size_bits++;
	/*
	get_code_string(node, aux_str);
	printf("node code after concatenate bit: '%s' (byte=%d, bit=%d)\n", aux_str, byte_position, bit_position);
	*/
	code_node(0, node->code, node->code_size_bits, node->left_node);
	code_node(1, node->code, node->code_size_bits, node->right_node);
}

void code_node_from_root(t_node *root) {
	if (root == NULL)
		return;
	code_node(0, root->code, 0, root->left_node);
	code_node(1, root->code, 0, root->right_node);
}

void show_node(int level, t_node *node) {
	int i;
	char buffer[80];

	if (node == NULL)
		return;
	for (i = 0; i < level; i++)
		printf("   ");
	get_code_string(node, buffer);
	if (node->data < 0)
		printf(" *  (%d) [%s]\n", node->freq, buffer);
	else {
		if (isgraph(node->data))
			printf("'%c' (%d) [%s]\n", node->data, node->freq, buffer);
		else
			printf("%02x  (%d) [%s]\n", node->data, node->freq, buffer);
	}
	show_node(level + 1, node->left_node);
	show_node(level + 1, node->right_node);
}

void show_list() {
	int i;
	t_node *aux;

	for (i = 0; i < list_size; i++) {
		aux = list[i];
		show_node(0, aux);
	}
	printf("\n");
}

unsigned short save_file_node_header_array(int fd, t_node *root) {
	t_file_node_header *file_node = malloc(sizeof(t_file_node_header));
	unsigned short offset;
	int byte_size;

	file_node->data = (root->data >= 0) ? ((unsigned char) root->data) : 0;
	if (root->left_node != NULL)
		file_node->left_node_offset = save_file_node_header_array(fd, root->left_node);
	else
		file_node->left_node_offset = 0;
	if (root->right_node != NULL)
		file_node->right_node_offset = save_file_node_header_array(fd, root->right_node);
	else
		file_node->right_node_offset = 0;
	offset = (unsigned short) lseek(fd, 0, SEEK_CUR);
	write(fd, file_node, sizeof(t_file_node_header));
	//printf("D: offset=%X (d='%c' [%02x], l=%u, r=%u)\n", offset, (isgraph(file_node->data)) ? file_node->data : ' ', file_node->data, file_node->left_node_offset, file_node->right_node_offset);
	return offset;
}

void save_compressed(int fd, t_node *root, unsigned char *compressed_data, int compressed_size_bits) {
	unsigned char buffer[80];
	unsigned short root_node_offset;
	unsigned short aux;
	unsigned short data_offset;
	int byte_position, bit_position, i;

	// signature
	memcpy(buffer, "HUF", 3);
	write(fd, buffer, 3);
	// root node offset (dummy value)
	write(fd, buffer, 2);
	// data beginning offset (dummy value)
	write(fd, buffer, 2);
	// node list
	root_node_offset = save_file_node_header_array(fd, root);
	// rewind to write REAL root node offset
	lseek(fd, 3, SEEK_SET);
	write(fd, &root_node_offset, sizeof(unsigned short));
	// goto EOF to continue
	lseek(fd, 0, SEEK_END);
	// begin data section
	memcpy(buffer, "DATA", 4);
	write(fd, buffer, 4);
	// rewind to write REAL data beginning offset
	data_offset = lseek(fd, 0, SEEK_CUR);
	lseek(fd, 5, SEEK_SET);
	write(fd, &data_offset, sizeof(unsigned short));
	// goto EOF to continue
	lseek(fd, 0, SEEK_END);
	// write compressed size (bits)
	write(fd, &compressed_size_bits, sizeof(int));
	// write the bits
	i = compressed_size_bits >> 3;
	if ((compressed_size_bits & 7) > 0)
		i++;
	write(fd, compressed_data, i);
}

int main(int argc, char *argv[]) {
	t_node *aux;
	unsigned char *source_data;
	int source_data_size;
	unsigned char *dest_data;
	int dest_data_size_bits;
	int dest_data_size;
	unsigned char *source_again_data;
	int source_again_data_size;
	int source_fd;
	int dest_fd;

	if (argc > 2) {
		// open source file and read source data
		if ((source_fd = open(argv[1], O_RDONLY)) < 0) {
			perror("open source file");
			return -1;
		}
		source_data_size = lseek(source_fd, 0, SEEK_END);
		lseek(source_fd, 0, SEEK_SET);
		source_data = (unsigned char *) malloc(source_data_size);
		read(source_fd, source_data, source_data_size);
		close(source_fd);
		// open destination file
		if ((dest_fd = open(argv[2], O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0) {
			perror("open destination file");
			return -1;
		}
	}
	else {
		fprintf(stderr, "syntax: %s source_file destination_file\n", argv[0]);
		return -1;
	}
	//printf("compressing '%s' (size=%d)...\n", source_data, source_data_size);
	read_list(source_data, source_data_size);
	order_list(0, list_size - 1);
	while (list_size > 1) {
		aux = join_first_two_nodes();
		add_node(aux);
		order_list(0, list_size - 1);
	}
	code_node_from_root(list[0]);
	show_list();
	compress(list[0], source_data, source_data_size, &dest_data, &dest_data_size_bits);


	printf("saving compressed data...\n");
	save_compressed(dest_fd, list[0], dest_data, dest_data_size_bits);
	close(dest_fd);


	dest_data_size = dest_data_size_bits >> 3;
	if ((dest_data_size_bits & 7) != 0)
		dest_data_size++;
	printf("compressed data size: %d bits (%d bytes)\n", dest_data_size_bits, dest_data_size);
	return 0;
}
